;; Generated package description from ftable.el  -*- no-byte-compile: t -*-
(define-package "ftable" "1.1" "Fill a table to fit in n columns" '((emacs "26.0")) :commit "d7c243ac3be2a67986f53f3dc31ba1426e82b771" :authors '(("Yuan Fu" . "casouri@gmail.com")) :maintainer '("Yuan Fu" . "casouri@gmail.com") :keywords '("convenience" "text" "table") :url "https://github.com/casouri/ftable")
